﻿// ExBox Tester application Copyright (c) 2011 by David Ek. All rights reserved.
//
// Freely usable and distributable for strictly noncommmercial purposes.

using System;
using System.IO.Ports;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace EkBoxTester
{
   public partial class MainForm : Form
   {
      #region Members

      /// <summary>
      /// Resolution of azimuth encoder. Valid range 2 to 65535. Value is zero if the text
      /// in azTextBox.Text is validated and found to be invalid.
      /// </summary>
      private int m_iAzRes;

      /// <summary>
      /// Resolution of altitude encoder. Valid range 2 to 65535. Value is zero if the text
      /// in altTextBox.Text is validated and found to be invalid.
      /// </summary>
      private int m_iAltRes;

      /// <summary>
      /// Name of the serial port to which the EkBox is connected (e.g. "COM1"). Is an empty string
      /// if the text in portComboBox is not a valid serial port name.
      /// </summary>
      private string m_sPort;

      /// <summary>
      /// The SerialPort object used to communicate with the EkBox.
      /// </summary>
      private SerialPort m_serialPort = null;

      #endregion Members

      #region Constructors

      /// <summary>
      /// the default form constructor
      /// </summary>
      public MainForm()
      {
         InitializeComponent();
      }

      #endregion Constructors

      #region EventHandlers

      /// <summary>
      /// Called when the form is closed. Make sure the serial port gets closed.
      /// </summary>
      /// <param name="sender">not used</param>
      /// <param name="e">not used</param>
      private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
      {
         StopTest();
      }

      /// <summary>
      /// MainForm_Load is called when the form is first loaded (the application is started)
      /// </summary>
      /// <param name="sender">not used</param>
      /// <param name="e">not used</param>
      private void MainForm_Load(object sender, EventArgs e)
      {
         this.Text = "EkBox Tester Version " + System.Reflection.Assembly.GetExecutingAssembly().GetName().Version.ToString();

         // initialize the serial port name combo box with the list of available ports. Note that
         // the combo box is of style "DropDown" to allow a port name to be manually typed in. This is to
         // handle the issue with some bluetooth devices not properly writing their port names to the registry,
         // resulting in extraneous characters in the name.

         string[] sPortNames = System.IO.Ports.SerialPort.GetPortNames();
         foreach (string sPort in sPortNames)
         {
            portComboBox.Items.Add(sPort);
         }

         // set initial values of input controls to valid values:

         portComboBox.SelectedIndex = 0; // select the first item in the list
         m_sPort = portComboBox.Text;
         azTextBox.Text = "4000";
         m_iAzRes = 4000;
         altTextBox.Text = "4000";
         m_iAltRes = 4000;

         // Initialize the state of the buttons, labels, and timers to pre-test condition.

         ResetDisplay();

         return;
      }

      /// <summary>
      /// altTextBox_Validated is called after the altTextBox control loses focus. We validate the value
      /// in the control and save it to m_iAltRes if it's valid.
      /// </summary>
      /// <param name="sender">not used</param>
      /// <param name="e">not used</param>
      private void altTextBox_Validated(object sender, EventArgs e)
      {
         // check the alt resolution input. Set the ErrorProvider if it's not valid.
         // IsValidResolution returns 0 for m_iAltRes if the input is not valid. Otherwise, it
         // returns the value entered.

         if (!IsValidResolution(altTextBox.Text, out m_iAltRes))
         {
            altResErrorProvider.SetError(altTextBox, "Alt Resolution must be between 2 and 65535");
         }
         else
         {
            altResErrorProvider.SetError(altTextBox, String.Empty);
         }

         return;
      }

      /// <summary>
      /// azTextBox_Validated is called after the azTextBox control loses focus. We validate the value
      /// in the control and save it to m_iAzRes if it's valid.
      /// </summary>
      /// <param name="sender">not used</param>
      /// <param name="e">not used</param>
      private void azTextBox_Validated(object sender, EventArgs e)
      {
         // check the az resolution input. Set the ErrorProvider if it's not valid.
         // IsValidResolution returns 0 for m_iAztRes if the input is not valid. Otherwise, it
         // returns the value entered.

         if (!IsValidResolution(azTextBox.Text, out m_iAzRes))
         {
            azResErrorProvider.SetError(azTextBox, "Az Resolution must be between 2 and 65535");
         }
         else
         {
            azResErrorProvider.SetError(azTextBox, String.Empty);
         }

         return;
      }

      /// <summary>
      /// Called when Errors button is clicked. Gets the number of encoder errors
      /// from the EkBox and displays the number in the errorsLabel label.
      /// </summary>
      /// <param name="sender">not used</param>
      /// <param name="e">not used</param>
      private void errorsButton_Click(object sender, EventArgs e)
      {
         try
         {
            int iErrCount;

            GetEncoderErrors(out iErrCount);
            errorsLabel.Text = iErrCount.ToString();

            return;
         }
         catch (Exception ex)
         {
            StopTest();
            MessageBox.Show("Failed to get encoder error count. Make sure the EkBox is connected and powered up."
               + Environment.NewLine + Environment.NewLine + "Exception text: " + ex.Message,
               "EkBox Tester Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
         }
      }

      // check to make sure the serial port name has a valid format
      /// <summary>
      /// portComboBox_Validated is called when the portComboBox control loses focus. Check the
      /// selected/entered value and save it to m_sPort if it's valid.
      /// </summary>
      /// <param name="sender">not used</param>
      /// <param name="e">not used</param>
      private void portComboBox_Validated(object sender, EventArgs e)
      {
         // Expecting an entry of "COMxx", where xx is one or two digits.
         // Use regular expressions to see if "com" plus one or more digits
         // exists at the beginning of the string. If so, save what's found to m_sPort.
         // If nothing is found, set m_sPort to be empty and set the ErrorProvider.

         Regex regex = new Regex("(?<port>com\\d+)", RegexOptions.IgnoreCase);
         Match mt = regex.Match(portComboBox.Text);
         if (mt.Success)
         {
            m_sPort = mt.Groups["port"].Value.ToUpper();
            portComboBox.Text = m_sPort;
            portErrorProvider.SetError(portComboBox, String.Empty);
         }
         else
         {
            m_sPort = String.Empty;
            portErrorProvider.SetError(portComboBox, "Serial port name must be \"COMxx\" where xx is one or two digits.");
         }

         // stop any ongoing test and disconnect from the current serial port.
         StopTest();

         return;
      }

      /// <summary>
      /// Called with Get Resolutions button is cicked. Reads the encoder resolutions
      /// from the EkBox and displays them in the altResLabel and azResLabel labels.
      /// </summary>
      /// <param name="sender">not used</param>
      /// <param name="e">not used</param>
      private void resolutionsButton_Click(object sender, EventArgs e)
      {
         try
         {
            int iAzRes, iAltRes;

            GetEncoderResolutions(out iAzRes, out iAltRes);
            azResLabel.Text = iAzRes.ToString();
            altResLabel.Text = iAltRes.ToString();

            return;
         }
         catch (Exception ex)
         {
            StopTest();
            MessageBox.Show("Failed to get encoder resolutions. Make sure the EkBox is connected and powered up."
               + Environment.NewLine + Environment.NewLine + "Exception text: " + ex.Message,
               "EkBox Tester Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
         }
      }

      /// <summary>
      /// Set the encoder resolutions in the Ekbox. Check the values first to make sure they're valid.
      /// </summary>
      /// <param name="sender"></param>
      /// <param name="e"></param>
      private void setResolutionsButton_Click(object sender, EventArgs e)
      {
         // check the input az encoder resolution.

         if (m_iAzRes == 0)
         {
            MessageBox.Show("Az resolution is invalid. Please enter a value between 2 and 65535.", "EkBox Tester Error",
               MessageBoxButtons.OK, MessageBoxIcon.Error);
            azTextBox.Focus();

            return;
         }

         // check the input alt encoder resolution.

         if (m_iAltRes == 0)
         {
            MessageBox.Show("Alt resolution is invalid. Please enter a value between 2 and 65535.", "EkBox Tester Error",
               MessageBoxButtons.OK, MessageBoxIcon.Error);
            altTextBox.Focus();

            return;
         }

         try
         {
            // now initialize the encoders:

            Cursor.Current = Cursors.WaitCursor;
            SetResolutions();
            Cursor.Current = Cursors.Default;
         }
         catch (Exception ex)
         {
            Cursor.Current = Cursors.Default;
            StopTest();
            MessageBox.Show("Failed to set encoder resolutions. Make sure you have selected the correct serial port, "
               + "and make sure the EkBox is connected and powered up."
               + Environment.NewLine + Environment.NewLine + "Exception text: " + ex.Message,
               "EkBox Tester Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            return;
         }
      }

      /// <summary>
      /// startButton_Click is called when the Start Test button is clicked. 
      /// Open the serial port and turn on the timer so that the
      /// encoder positions are repeatedly read and displayed.
      /// </summary>
      /// <param name="sender">not used</param>
      /// <param name="e">not used</param>
      private void startButton_Click(object sender, EventArgs e)
      {
         // attempt to open the serial port. If not successful, return and abort.
         if (!Connect())
         {
            return;
         }

         // time to start the test. Disable the "Start Test" button and enable the "Stop Test" button. 

         startButton.Enabled = false;
         stopButton.Enabled = true;

         // enable the timer so that the encoder positions start getting read and displayed.

         timer.Enabled = true;

         return;
      }

      /// <summary>
      /// stopButton_Click is called when the Stop Test button is clicked. Stop the test.
      /// </summary>
      /// <param name="sender">not used</param>
      /// <param name="e">not used</param>
      private void stopButton_Click(object sender, EventArgs e)
      {
         StopTest();

         return;
      }

      /// <summary>
      /// timer_Tick is called every 200 ms while the timer is active, to read
      /// the encoder positions and update the display.
      /// </summary>
      /// <param name="sender">not used</param>
      /// <param name="e">not used</param>
      private void timer_Tick(object sender, EventArgs e)
      {
         int iAzTicks, iAltTicks; // the encoder position values

         try
         {
            // disable the timer while we're reading the encoder positions.

            timer.Enabled = false;

            // get the encoder positions.

            GetEncoderPositions(out iAzTicks, out iAltTicks);

            // format and display the encoder positions.

            readoutLabel.Text = String.Format("{0:00000}", iAzTicks)
               + Environment.NewLine
               + String.Format("{0:00000}", iAltTicks);
            
            // re-enable the timer for the next read.

            timer.Enabled = true;

            return;
         }

         // if any exceptions occur, stop the test.

         catch (Exception ex)
         {
            StopTest();
            MessageBox.Show("Failed to get encoder positions. Make sure the EkBox is connected and powered up."
               + Environment.NewLine + Environment.NewLine + "Exception text: " + ex.Message,
               "EkBox Tester Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
         }
      }

      #endregion EventHandlers

      #region PrivateMethods

      /// <summary>
      /// Checks to see if the serial port is open and opens it if needed.
      /// </summary>
      /// <returns>true if port is open, or false if port couldn't be opened</returns>
      private bool Connect()
      {
         // if already connected, just return true.
         if (m_serialPort != null && m_serialPort.IsOpen)
         {
            return true;
         }

         // check the input serial port name.

         if (String.IsNullOrEmpty(m_sPort))
         {
            MessageBox.Show("Serial port name is invalid. Please select a port from the list, or enter a port name "
               + "in the form of \"COMxx\" where xx is one or two digits.", "EkBox Tester Error",
               MessageBoxButtons.OK, MessageBoxIcon.Error);
            portComboBox.Focus();

            return false;
         }

         // try to create and open the serial port:

         try
         {
            m_serialPort = new SerialPort(m_sPort, 9600, Parity.None, 8, StopBits.One);
            m_serialPort.Handshake = Handshake.None;
            m_serialPort.ReadTimeout = 5000;
            m_serialPort.WriteTimeout = 5000;
         }
         catch (Exception ex)
         {
            StopTest();
            MessageBox.Show("Could not create serial port object for port " + m_serialPort + "."
               + Environment.NewLine + Environment.NewLine + "Exception text: " + ex.Message,
               "EkBox Tester Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            return false;
         }

         try
         {
            m_serialPort.Open();
            m_serialPort.DiscardInBuffer();
            return true;
         }
         catch (Exception ex)
         {
            StopTest();
            MessageBox.Show("Could not open serial port " + m_serialPort + "."
               + Environment.NewLine + Environment.NewLine + "Exception text: " + ex.Message,
               "EkBox Tester Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            return false;
         }
      }

      /// <summary>
      /// Read the number of encoder errors from the EkBox.
      /// Any exceptions thrown should be caught by the calling method.
      /// </summary>
      /// <param name="a_iErrCount">the number of encoder errors occuring since the last check</param>
      private void GetEncoderErrors(out int a_iErrCount)
      {
         a_iErrCount = 0;

         // attempt to open the serial port. If not successful, return and abort.
         if (!Connect())
         {
            return;
         }

         bool timerEnabled = timer.Enabled;

         if (timerEnabled)
         {
            timer.Enabled = false;
         }

         // send a 'p' (single byte) to the EkBox. The EkBox should reply with
         // one byte indicating the number of encoder errors occurring since the
         // previous 'p' command was processed.

         byte[] p = new byte[1] { (byte)'p' }; // the command to be sent

         // send the command.

         m_serialPort.Write(p, 0, 1);

         // get the reply.

         a_iErrCount = m_serialPort.ReadByte(); // the number of errors

         if (timerEnabled)
         {
            timer.Enabled = true;
         }

         return;
      }

      /// <summary>
      /// Read the encoder positions from the EkBox.
      /// Any exceptions thrown should be caught by the calling method.
      /// </summary>
      /// <param name="a_iAzTicks">the position of the az encoder</param>
      /// <param name="a_iAltTicks">the position of the alt encoder</param>
      private void GetEncoderPositions(out int a_iAzTicks, out int a_iAltTicks)
      {
         a_iAltTicks = 0;
         a_iAzTicks = 0;

         // attempt to open the serial port. If not successful, return and abort.
         if (!Connect())
         {
            return;
         }

         bool timerEnabled = timer.Enabled;

         if (timerEnabled)
         {
            timer.Enabled = false;
         }

         // send a 'y' (single byte) to the EkBox. The EkBox should reply with
         // four bytes (two two-byte unsigned integers).

         byte[] p = new byte[1] { (byte)'y' }; // the command to be sent
         int[] result = new int[4]; // int array to hold the reply

         // send the command.

         m_serialPort.Write(p, 0, 1);

         // get the reply.

         result[0] = m_serialPort.ReadByte(); // low byte of altitude integer
         result[1] = m_serialPort.ReadByte(); // high byte of altitude integer
         result[2] = m_serialPort.ReadByte(); // low byte of azimuth integer
         result[3] = m_serialPort.ReadByte(); // high byte of azimuth integer

         // convert the bytes to the integer positions.

         a_iAltTicks = result[0] + 256 * result[1];
         a_iAzTicks = result[2] + 256 * result[3];

         if (timerEnabled)
         {
            timer.Enabled = true;
         }

         return;
      }

      /// <summary>
      /// Read the encoder resolutions from the EkBox.
      /// Any exceptions thrown should be caught by the calling method.
      /// </summary>
      /// <param name="a_iAzRes">the resolution of the az encoder</param>
      /// <param name="a_iAltRes">the resolution of the alt encoder</param>
      private void GetEncoderResolutions(out int a_iAzRes, out int a_iAltRes)
      {
         a_iAltRes = 0;
         a_iAzRes = 0;

         // attempt to open the serial port. If not successful, return and abort.
         if (!Connect())
         {
            return;
         }

         bool timerEnabled = timer.Enabled;

         if (timerEnabled)
         {
            timer.Enabled = false;
         }

         // send a 'h' (single byte) to the EkBox. The EkBox should reply with
         // four bytes (two two-byte unsigned integers).

         byte[] p = new byte[1] { (byte)'h' }; // the command to be sent
         int[] result = new int[4]; // int array to hold the reply

         // send the command.

         m_serialPort.Write(p, 0, 1);

         // get the reply.

         result[0] = m_serialPort.ReadByte(); // low byte of altitude resolution integer
         result[1] = m_serialPort.ReadByte(); // high byte of altitude resolution integer
         result[2] = m_serialPort.ReadByte(); // low byte of azimuth resolution integer
         result[3] = m_serialPort.ReadByte(); // high byte of azimuth resolution integer

         // convert the bytes to the integer positions.

         a_iAltRes = result[0] + 256 * result[1];
         a_iAzRes = result[2] + 256 * result[3];

         if (timerEnabled)
         {
            timer.Enabled = true;
         }

         return;
      }

      /// <summary>
      /// Check the input encoder resolution string for validity, and return its value
      /// as an integer
      /// </summary>
      /// <param name="a_sVal">the input encoder resolution string</param>
      /// <param name="a_iVal">the integer value of the input, or 0 if the input is not valid</param>
      /// <returns>true if the input string is valid, otherwise false</returns>
      private bool IsValidResolution(string a_sVal, out int a_iVal)
      {
         int iVal;

         a_iVal = 0; // default value for the output if the input string is invalid.

         // is the string empty?

         if (String.IsNullOrWhiteSpace(a_sVal))
         {
            return false;
         }

         // does it not parse as an integer?

         if (!Int32.TryParse(a_sVal, out iVal))
         {
            return false;
         }

         // is its value in the valid range?

         if ((iVal < 2) || (iVal > 65535))
         {
            return false;
         }

         // input string is valid. Return its integer value in the output.

         a_iVal = iVal;

         return true;
      }

      /// <summary>
      /// Resets the input controls, buttons, labels, and timer for pre-test conditions.
      /// </summary>
      private void ResetDisplay()
      {
         timer.Enabled = false;
         startButton.Enabled = true;
         stopButton.Enabled = false;

         return;
      }
      
      /// <summary>
      /// Send the encoder resolutions to the EkBox.
      /// Any exceptions should be caught by calling method.
      /// </summary>
      private void SetResolutions()
      {
         // attempt to open the serial port. If not successful, return and abort.
         if (!Connect())
         {
            return;
         }

         bool timerEnabled = timer.Enabled;

         if (timerEnabled)
         {
            timer.Enabled = false;
         }

         // send 'z' followed by the alt and az resolutions.
         // the alt and az resolutions are each sent as a two-byte unsigned integer, low byte
         // then high byte.

         byte[] p = new byte[5];

         p[0] = (byte)'z';
         p[1] = (byte)(m_iAltRes % 256); // low byte of alt res
         p[2] = (byte)(m_iAltRes / 256); // high byte of alt res
         p[3] = (byte)(m_iAzRes % 256);  // low byte of az res
         p[4] = (byte)(m_iAzRes / 256);  // high byte of az res

         m_serialPort.Write(p, 0, 5);

         // EkBox should respond with 'r'. If it doesn't, a timeout exception gets thrown.

         int reply = m_serialPort.ReadByte();
         string result = ((char)reply).ToString();

         if (timerEnabled)
         {
            timer.Enabled = true;
         }

         return;
      }

      /// <summary>
      /// StopTest is called to stop reading the encoder positions, close the serial port, and reset
      /// the state of the controls to pre-test condition.
      /// </summary>
      private void StopTest()
      {
         if (m_serialPort != null && m_serialPort.IsOpen)
         {
            m_serialPort.Close();
         }

         ResetDisplay();

         return;
      }

      #endregion PrivateMethods
   }
}
